# encoding: UTF-8
require 'tomte/document/keys/key'

module Tomte
  module Document
    module Keys
      extend ActiveSupport::Concern

      included do
        extend ActiveSupport::DescendantsTracker
        key :_id, Tomte::ObjectId, :default => lambda { Tomte::ObjectId.new }
      end

      module ClassMethods
        def inherited(descendant)
          descendant.instance_variable_set(:@keys, keys.dup)
          super
        end

        def keys
          @keys ||= {}
        end

        def key(*args)
          Key.new(*args).tap do |key|
            keys[key.name] = key
            create_accessors_for(key)
            create_key_in_descendants(*args)
          end
        end

        def key?(key)
          keys.keys.include?(key.to_s)
        end

        def using_object_id?
          object_id_key?(:_id)
        end

        def object_id_keys
          keys.keys.select { |key| keys[key].type == Tomte::ObjectId }.map { |k| k.to_sym }
        end

        def object_id_key?(name)
          object_id_keys.include?(name.to_sym)
        end

        def to_tomte(instance)
          return nil if instance.nil?
          instance.to_tomte
        end

        def from_tomte(value)
          return nil if value.nil?
          value.is_a?(self) ? value : load(value)
        end

        def load(attrs)
          return nil if attrs.nil?
          res =
            begin
              attrs['_doctype'].present? ? attrs['_doctype'].constantize : self
            rescue NameError
              self
            end.allocate.initialize_from_data(attrs)
          res.__send__(:document_initialize, attrs) if res.respond_to?(:document_initialize)
          res
        end

        private
        def key_accessors_module_defined?
          if method(:const_defined?).arity == 1 # Ruby 1.9 compat check
            const_defined?('TomteDocumentKeys')
          else
            const_defined?('TomteDocumentKeys', false)
          end
        end

        def accessors_module
          if key_accessors_module_defined?
            const_get 'TomteDocumentKeys'
          else
            const_set 'TomteDocumentKeys', Module.new
          end
        end

        def create_accessors_for(key)
          accessors_module.module_eval <<-end_eval
            def #{key.name}
              read_key(:#{key.name})
            end

            def #{key.name}_before_typecast
              read_key_before_typecast(:#{key.name})
            end

            def #{key.name}=(value)
              write_key(:#{key.name}, value)
            end

            def #{key.name}?
              read_key(:#{key.name}).present?
            end
          end_eval

          include accessors_module
        end

        def create_key_in_descendants(*args)
          descendants.each { |descendant| descendant.key(*args) }
        end
      end

      def initialize(attrs={})
        @_new = true
        self.attributes = attrs
      end

      def initialize_from_data(attrs={})
        @_new = false
        load_from_data(attrs)
        self
      end

      def attributes=(attrs)
        return if attrs.blank?

        attrs.each_pair do |key, value|
          if respond_to?(:"#{key}=")
            self.send(:"#{key}=", value)
          else
            self[key] = value
          end
        end
      end

      def attributes
        HashWithIndifferentAccess.new.tap do |attrs|
          keys.select { |name, key| !self[key.name].nil? || key.type == ObjectId }.each do |name, key|
            value = key.set(self[key.name])
            attrs[name] = value
          end

          #embedded_associations.each do |association|
          #  if documents = instance_variable_get(association.ivar)
          #    if association.one?
          #      attrs[association.name] = documents.to_tomte
          #    else
          #      attrs[association.name] = documents.map { |document| document.to_tomte }
          #    end
          #  end
          #end
        end
      end
      #alias :to_tomte :attributes

      def to_tomte
        HashWithIndifferentAccess.new.tap do |attrs|
          keys.select { |name, key| !self[key.name].nil? || key.type == ObjectId }.each do |name, key|
            value =
              (self[key.name].respond_to?(:to_tomte) ? self[key.name].to_tomte : key.set(self[key.name]))
            attrs[name] = value
          end
        end
      end


      def update_attributes(attrs={})
        assign(attrs)
      end

      def update_attributes!(attrs={})
        assign(attrs)
      end

      def update_attribute(name, value)
        update_attributes(name.to_sym => value)
      end

      def assign(attrs={})
        self.attributes = attrs
      end

      def id
        _id
      end

      def id=(value)
        if self.class.using_object_id?
          value = ObjectId.to_to_tomte(value)
        end

        self[:_id] = value
      end

      def [](name)
        read_key(name)
      end

      def []=(name, value)
        ensure_key_exists(name)
        write_key(name, value)
      end

      def keys
        self.class.keys
      end

      def key_names
        keys.keys
      end

      def new?
        @_new ||= true
      end

      private
      def load_from_data(attrs)
        return if attrs.blank?
        attrs.each do |key, value|
          if respond_to?(:"#{key}=") && !self.class.key?(key)
            self.send(:"#{key}=", value)
          else
            self[key] = value
          end
        end
      end

      def ensure_key_exists(name)
        self.class.key(name) unless respond_to?("#{name}=")
      end

      def read_key(key_name)
        if key = keys[key_name.to_s]
          value = key.get(instance_variable_get(:"@#{key_name}"))
          instance_variable_set(:"@#{key_name}", value)
        end
      end

      def read_key_before_typecast(name)
        instance_variable_get(:"@#{name}_before_typecast")
      end

      def write_key(name, value)
        key = keys[name.to_s]
        casted_value = key.try_cast(value)

        instance_variable_set :"@#{name}_before_typecast", casted_value
        instance_variable_set :"@#{name}", key.set(casted_value)
      end

    end
  end
end

