
# Tomte, a framework for distributed agents.
# Copyright (c) 2010-2013, Lele Forzani, windmillmedia. All Rights Reserved.
#
# Author:: Lele Forzani <lele@windmill.it>
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.

require 'fileutils'

# Defines the log level for a Tomte project.
TOMTE_LOG_LEVEL = ENV['TOMTE_LOG_LEVEL'] unless defined?(TOMTE_LOG_LEVEL)
# Defines the logger used for a Tomte project.
TOMTE_LOGGER_CONFIG = ENV['TOMTE_LOGGER_CONFIG'] unless defined?(TOMTE_LOGGER_CONFIG)

module Tomte

#  def self.logger
#    Tomte::Logger.new(config.core.merge(config.local)) if Thread.current[:_logger].nil?
#    Thread.current[:_logger]
#  end

#  def self.logger=(value)
#    Thread.current[:_logger] = value
#  end
  
  #
  # Extensions to the built in Ruby logger.
  #
  class Logger
    include Logging::LoggingExtensions

    attr_accessor :level
    attr_accessor :auto_flush
    attr_reader   :buffer
    attr_reader   :log
    attr_reader   :init_args

    attr_accessor :identity
    attr_accessor :environment

    @@mutex = {}

    #
    # Configuration for a given environment, possible options are:
    #
    # :log_level:: One of [:fatal, :error, :warn, :info, :debug]
    # :stream:: One of [:to_file, :null, :stdout, :stderr] a string representing a path name
    #   our your custom stream
    # :log_level::
    #   The log level from, e.g. :fatal or :info. Defaults to :warn in the
    #   production environment and :debug otherwise.
    # :auto_flush::
    #   Whether the log should automatically flush after new messages are
    #   added. Defaults to true.
    # :format_datetime:: Format of datetime. Defaults to: "%d/%b/%Y %H:%M:%S"
    # :format_message:: Format of message. Defaults to: ""%s - - [%s] \"%s\"""
    #
    # @example
    #   Logger::Config[:development] = { :log_level => :debug, :stream => :to_file }
    #   # or you can edit our defaults
    #   Logger::Config[:development][:log_level] = :error
    #   # or you can use your stream
    #   Logger::Config[:development][:stream] = StringIO.new
    #
    # Defaults are:
    #
    #   :production  => { :log_level => :warn,  :stream => :to_file },
    #   :daemonized  => { :log_level => :info, :stream => :stdout },
    #   :interactive => { :log_level => :debug, :stream => :stdout, :format_datetime => ' ' },
    #   :development => { :log_level => :debug, :stream => :stdout, :format_datetime => ' ' },
    #   :unknown     => { :log_level => :debug, :stream => :stdout, :format_datetime => ' ' },
    #   :test        => { :log_level => :debug, :stream => :null }
    #
    # In some cases, configuring the loggers before loading the framework is necessary.
    # You can do so by setting TOMTE_LOGGER:
    #
    #   TOMTE_LOGGER = { :staging => { :log_level => :debug, :stream => :to_file }}
    #

    Config = {
      :production  => { :color => true,  :log_level => :warn,  :stream => :to_file },
      :daemonized  => { :color => false, :log_level => :info,  :stream => :to_file },
      :default     => { :color => true,  :log_level => :info,  :stream => :stdout, :format_datetime => ' ' },
      :debug       => { :color => true,  :log_level => :debug, :stream => :stdout, :format_datetime => ' ' },
      :interactive => { :color => true,  :log_level => :warn,  :stream => :stderr, :format_datetime => ' ' },
      :nocolor     => { :color => false, :log_level => :info,  :stream => :stdout, :format_datetime => ' ' },
      :unknown     => { :color => false, :log_level => :debug, :stream => :stdout, :format_datetime => ' ' },
      :test        => { :color => true,  :log_level => :debug, :stream => :null }
    }
    Config.merge!(TOMTE_LOGGER_CONFIG) if  (defined?(TOMTE_LOGGER_CONFIG) && !TOMTE_LOGGER_CONFIG.nil?)
    


    def self.setup!(options)

      environment, identity = [(Tomte.env || :test).to_sym, (options[:identity] || 'unknown')]
      profile = options[:profile].to_sym || environment
      options.delete(:log_level) if options[:log_level].nil?

      config = Config[profile]

      unless config
        warn("No logging configuration for :#{profile} found, falling back to :unknown")
        config = Config[:unknown]
      end

      stream = case config[:stream]
        when String
          dir = File.dirname(config[:stream])
          FileUtils.mkdir_p(dir) unless File.exists?(dir)
          File.new(config[:stream], "a+")
        when :to_file
          raise ArgumentError, "Log path not given." if options[:log_path].blank?
          FileUtils.mkdir_p(options[:log_path])
          File.new(File.join(options[:log_path], "%s-%s.log" % [identity, environment]), "a+")
        when :null   then StringIO.new
        when :stdout then $stdout
        when :stderr then $stderr
        else config[:stream]
      end

      Thread.current[:_logger] = new(config.merge(options.merge(:stream => stream, :environment => environment)))
    end


    ##
    # To initialize the logger you create a new object, proxies to set_log.
    #
    # @param [Hash] options
    #
    # @option options [Symbol] :stream ($stdout)
    #   Either an IO object or a name of a logfile. Defaults to $stdout
    #
    # @option options [Symbol] :log_level (:production in the production environment and :debug otherwise)
    #   The log level from, e.g. :fatal or :info.
    #
    # @option options [Symbol] :auto_flush (true)
    #   Whether the log should automatically flush after new messages are
    #   added. Defaults to true.
    #
    # @option options [Symbol] :format_datetime (" [%d/%b/%Y %H:%M:%S] ")
    #   Format of datetime
    #
    # @option options [Symbol] :format_message ("%s -%s%s")
    #    Format of message
    #
    # @option options [Symbol] :log_static (false)
    #   Whether or not to show log messages for static files.
    #
    def initialize(options={})

      @buffer          = []
      @color           = options[:color]
      @auto_flush      = options.has_key?(:auto_flush) ? options[:auto_flush] : true
      @level           = options[:log_level] ? Logging::Levels[options[:log_level].to_sym] : Logging::Levels[:debug]
      @log             = options[:stream]  || $stdout
      @log.sync        = true
      @mutex           = @@mutex[@log] ||= Mutex.new
      @format_datetime = options[:format_datetime] || "%d/%b/%Y %H:%M:%S"
      @format_message  = options[:format_message]  || "%s - %s %s"

      @identity = options[:identity] || 'unknown'
      @environment = options[:environment] || Tomte.env || 'unknown'

      colorize! if @color

      tomte_debug do
        puts "== logger configuration ==".yellow.bold
        ap options
      end

      debug "+ Logger for '#{identity}:#{environment}' started, level is #{level}."
    end


    ##
    # Changes the log level by symbol
    #
    # @param [String, Symbol] level
    #
    #
    def set_level(loglevel)
      @level =  Logging::Levels[loglevel]
    end

    ##
    # Adds a message to the log - for compatibility with other loggers.
    #
    def add(level, message = nil)
      write(message)
    end


    ##
    # Flush the entire buffer to the log object.
    #
    def flush
      return unless @buffer.size > 0
      @mutex.synchronize do
        @log.write(@buffer.slice!(0..-1).join(''))
      end
    end

    ##
    # Close and remove the current log object.
    #
    # @return [NilClass]
    #
    def close
      flush
      @log.close if @log.respond_to?(:close) && !@log.tty?
      @log = nil
    end

    ##
    # Directly append message to the log.
    #
    # @param [String] message
    #   The message
    #
    def <<(message = nil)
      message << "\n" unless message[-1] == ?\n
      @buffer << message
      flush if @auto_flush
      message
    end
    alias :write :<<

    ##
    # Formats the log.
    #
    def format(message, level)
      @format_message % [stylized_level(level), colorize(Time.now.strftime(@format_datetime), :yellow), filter(message.to_s.strip)]
    end

  end
end


