module Tomte

  module Logging
    Levels = {
      :fatal =>  ::Logger::FATAL,
      :error =>  ::Logger::ERROR,
      :warn  =>  ::Logger::WARN,
      :info  =>  ::Logger::INFO,
      :debug =>  ::Logger::DEBUG,
      :devel => -1,
    } unless const_defined?(:Levels)

    module LoggingExtensions
     ##
      # Generate the logging methods for {Tomte.logger} for each log level.
      #
      Tomte::Logging::Levels.each_pair do |name, number|
        define_method(name) do |*args|
          return if number < level
          if args.size > 1
            bench(*args)
          else
            push(args * '', name)
          end
        end

        define_method(:"#{name}?") do
          number >= level
        end
      end

      ##
      # Appends a message to the log. The methods yield to an optional block and
      # the output of this block will be appended to the message.
      #
      # @param [String] message
      #   The message that you want write to your stream
      #
      # @param [String] level
      #   The level one of :debug, :warn etc...
      #
      #
      def push(message = nil, level = nil)
        add(Levels[level], format(message, level))
      end

      ##
      # logs an exception
      #
      def exception(exception, ex_level = :error, silence_backtrace = false)
        return if Levels[ex_level] < level
        ex  = "==  #{exception.class}\n"
        ex += "    #{exception.class} (#{exception.message}):\n\n".subindent.important!
        ex += ("backtrace:\n" + exception.backtrace.join("\n")).subindent.blue unless
            (exception.backtrace.nil? || silence_backtrace)

        push(ex, ex_level)
      end

      ##
      # Append a to development logger a given action with time
      #
      # @param [string] action
      #   The action
      #
      # @param [float] time
      #   Time duration for the given action
      #
      # @param [message] string
      #   The message that you want to log
      #
      # @example
      #   logger.bench 'GET', started_at, '/blog/categories'
      #   # => DEBUG - GET (0.056ms) - /blog/categories
      #
      def bench(action, began_at, message, level=:debug, color=:yellow)
        @_pad  ||= 8
        @_pad    = action.to_s.size if action.to_s.size > @_pad
        duration = Time.now - began_at
        color    = :red if duration > 1
        push "%s (" % colorize(action.to_s.upcase.rjust(@_pad), color) + colorize("%0.4fms", :bold, color) % duration + ") %s" % message.to_s, level
      end


      ##
      # Formats the log message. This method is a noop and should be implemented by other
      # logger components such as {Tomte::Logger}.
      #
      # @param [String] message
      #   The message to format
      #
      # @param [String,Symbol] level
      #   The log level, one of :debug, :warn...
      def format(message, level)
        message
      end

      ##
      # The log level, with some style added. May be reimplemented.
      #
      # @example
      #   stylized_level(:debug) => DEBUG
      #
      # @param [String,Symbol] level
      #   The log level
      #
      def stylized_level(level)
        level.to_s.upcase.rjust(7)
      end

      ##
      # Colorizes a string for colored console output. This is a noop and can be reimplemented
      # to colorize the string as needed.
      #
      # @see
      #   ColorizedLogger
      #
      # @param [string]
      #   The string to be colorized.
      #
      # @param [Array<Symbol>]
      #   The colors to use. Should be applied in the order given.
      def colorize(string, *colors)
        string
      end

      ##
      # Filters the log string to remove color codes in case they are passed
      #
      # @see
      #   ColorizedLogger
      #
      # @param [string]
      #   The string to be filtered
      #
      def filter(string)
        string.unansi
      end

      ##
      # Turns a logger with LoggingExtensions into a logger with colorized output.
      #
      # @example
      #   Tomte.logger = Logger.new($stdout)
      #   Tomte.logger.colorize!
      #   Tomte.logger.debug("Fancy Tomte debug string")
      def colorize!
        self.extend(ColorizedLogger)
      end

    end

    module ColorizedLogger
    # Colors for levels
      ColoredLevels = {
        :fatal => [:red],
        :error => [:red],
        :warn  => [:yellow],
        :info  => [:green],
        :debug => [:cyan],
        :devel => [:magenta]
      } unless defined?(ColoredLevels)


      def filter(string)
        string
      end

      ##
      # Colorize our level
      #
      # @param [String, Symbol] level
      #
      # @see Logging::ColorizedLogger::ColoredLevels
      #
      def colorize(string, *colors)
        colors.each do |c|
          string = string.send(c)
        end
        string
      end

      def stylized_level(level)
        style = ColoredLevels[level].map { |c| "\e[%dm" % String.colors[c] } * ''
        [style, super, "\e[0m"] * ''
      end
    end
  end
end